from netgen.geom2d import unit_square
from ngsolve import (Mesh, Draw, Redraw, CoefficientFunction, sqrt, x, y, exp,
                     L2, GridFunction, TaskManager)
from ngsolve.internal import visoptions
from ngstents import TentSlab
from ngstents.conslaw import Euler

maxh = 0.05
mesh = Mesh(unit_square.GenerateMesh(maxh=maxh))

dt = 0.05
tend = 0.25

# using causality constant
local_ctau = True
global_ctau = 2/3
wavespeed = 6
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

order = 4
V = L2(mesh, order=order, dim=mesh.dim+2)
u = GridFunction(V,"u")
cl = Euler(u, ts, reflect=mesh.Boundaries("left|bottom|right|top"))
cl.SetTentSolver("SARK",substeps=2*order)

d = 5
rho = CoefficientFunction(0.1+exp(-200*((x-0.5)*(x-0.5)+(y-0.5)*(y-0.5))))
m = CoefficientFunction((0,0))
p = CoefficientFunction(0.1+exp(-200*((x-0.5)*(x-0.5)+(y-0.5)*(y-0.5))))
T = 2*p/rho
E = d/4*T*rho + 1/(2*rho)*m*m

cf = CoefficientFunction((rho,m,E))
cl.SetInitial(cf)
Draw(u)

visoptions.scalfunction = "u:1"
visoptions.vecfunction = None

t = 0
cnt = 0
with TaskManager():
    while t<tend-dt/2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt%1 == 0:
            print("{0:.5f}".format(t))
            Redraw(True)



__all__ = ["TentSlab", "Tent", "conslaw", "utils"]

import ngsolve
from ._pytents import TentSlab, Tent
from . import conslaw
from . import utils

from .utils._drawtents import DrawPitchedTentsPlt
from .utils._drawtents2d import DrawPitchedTents

TentSlab.DrawPitchedTentsPlt = DrawPitchedTentsPlt
TentSlab.DrawPitchedTents = DrawPitchedTents
del DrawPitchedTentsPlt, DrawPitchedTents



"""
Select a mesh vertex by number and draw each tent for which the
specified vertex is the tent vertex
"""

from ngstents import TentSlab
from netgen.csg import unit_cube
import ngsolve as ng
import matplotlib.pyplot as plt

mesh = ng.Mesh(unit_cube.GenerateMesh(maxh=0.3))
resp = input("Enter a mesh vertex number in 0:{}: ".format(mesh.nv-1))
try:
    vnr = int(resp)
except ValueError:
    vnr = -1
if 0 <= vnr < mesh.nv:
    tps = TentSlab(mesh)
    tps.SetMaxWavespeed(1.0)
    tps.PitchTents(dt=0.5)
    ntents = tps.GetNTents()
    tents = [(i, tps.GetTent(i)) for i in range(ntents)]
    tents = sorted([(i, t) for (i, t) in tents
                    if t.vertex == vnr], key=lambda pr: pr[1].level)
    plt.ion()
    for i, _ in tents:
        plt.close()
        ax = tps.Draw3DTentPlt(i)
        plt.show()
        input("press <Enter> ")
else:
    print("invalid vertex")



from netgen.geom2d import SplineGeometry
from ngsolve import Mesh, Draw, Redraw
from ngsolve import CoefficientFunction, sqrt, sin, cos, x, y
from ngsolve import InnerProduct, Id, L2, GridFunction, TaskManager, Integrate
from ngsolve import specialcf as scf
from ngsolve.internal import visoptions, viewoptions
from ngstents import TentSlab
from ngstents.conslaw import ConservationLaw
from math import pi
import time

geom = SplineGeometry()
geom.AddRectangle((0, 0), (pi, pi), bc="reflect")
ngmesh = geom.GenerateMesh(maxh=0.25)
mesh = Mesh(ngmesh)

# setting the problem
tend = 2 / sqrt(2) * pi
dt = tend / 10
wavespeed = 1

# using causality constant
local_ctau = True
global_ctau = 2 / 3
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

order = 2
V = L2(mesh, order=order, dim=mesh.dim + 1)
gfu = GridFunction(V, name="u")

# vector-valued coefficient function
n = scf.normal(mesh.dim)


def Flux(u):
    """
    Compute the flux for given TrialFunction u,
    a matrix-valued CoefficientFunction of size (h,w) where
    h represents the number of equations and
    w should correspond to the dimension of the mesh/space
    """
    return CoefficientFunction((Id(mesh.dim) * u[mesh.dim], u[0:mesh.dim]),
                               dims=(V.dim, mesh.dim))


def NumFlux(um, up):
    """
    Compute the numeric flux
    Inputs
    um: trace of u for current element on facet
    up: trace of u for other element on facet
    """
    # upwind flux
    flux = 0.5 * (Flux(um) + Flux(up)) * n
    flux_vec = 0.5 * (um[0:mesh.dim] - up[0:mesh.dim]) * n * n
    flux_scal = 0.5 * (um[mesh.dim] - up[mesh.dim])
    return flux + CoefficientFunction((flux_vec, flux_scal))


def InverseMap(y):
    """
    solves "y = u - (f(u),gradphi)" for u

    assuming wave speed = 1
    """
    norm_sqr = InnerProduct(ts.gradphi, ts.gradphi)
    mu = (y[mesh.dim] + InnerProduct(y[0:mesh.dim], ts.gradphi)) / (1 -
                                                                    norm_sqr)
    q = y[0:mesh.dim] + mu * ts.gradphi
    return CoefficientFunction((q, mu))


def BndNumFlux(um):
    """
    defines numerical flux on boundary elements using the normal vector n
    um: trace of u for current element on facet
    """
    # set (q,n) = 0
    return CoefficientFunction((um[mesh.dim] * n, 0))


cl = ConservationLaw(gfu,
                     ts,
                     flux=Flux,
                     numflux=NumFlux,
                     inversemap=InverseMap)
cl.SetBoundaryCF(mesh.BoundaryCF({"reflect": BndNumFlux(cl.u_minus)}))
# cl.SetTentSolver("SAT",stages=order+1, substeps=4*order)
cl.SetTentSolver("SARK", stages=order + 1, substeps=4 * order)

mu0 = cos(x) * cos(y)
q0 = CoefficientFunction(tuple(mesh.dim * [0]))
cl.SetInitial(CoefficientFunction((q0, mu0)))

Draw(gfu)
visoptions.scalfunction = "u:{:0}".format(mesh.dim + 1)
viewoptions.drawedges = 1

t = 0
cnt = 0
redraw = 1

# input("press enter to start")
t1 = time.time()
with TaskManager():
    while t < tend - dt / 2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt % redraw == 0:
            print("{:5f}".format(t))
            Redraw(True)
print("total time = ", time.time() - t1)

exsol = CoefficientFunction((sin(x) * cos(y) * sin(sqrt(2) * tend) / sqrt(2),
                             cos(x) * sin(y) * sin(sqrt(2) * tend) / sqrt(2),
                             cos(x) * cos(y) * cos(sqrt(2) * tend)))

Draw(exsol, mesh, "exact")
l2error = sqrt(
    Integrate(InnerProduct(gfu - exsol, gfu - exsol), mesh, order=3 * order))
print("l2error = ", l2error)



"""
Module test_causal_tents

Tests edge-based and volume-based methods in 1D, 2D and 3D

Note: In all tests, setting the global_ct slightly less than 1 ensures
      a causal slab.  If global_ct=1, the default, then roundoff error may
      cause tests to fail.
"""
from ngsolve import Mesh
from ngstents import TentSlab
from ngsolve.meshes import Make1DMesh


def Get1DMesh():
    return Make1DMesh(10)


def Get2DMesh():
    from netgen.geom2d import unit_square
    mesh = Mesh(unit_square.GenerateMesh(maxh=.3))
    mesh.Refine()
    return mesh


def Get3DMesh():
    from netgen.csg import unit_cube
    mesh = Mesh(unit_cube.GenerateMesh(maxh=.5))
    mesh.Refine()
    return mesh


def test_1D_vol_causal():
    mesh = Get1DMesh()
    dt = 10
    c = 1
    global_ct = 0.999
    method = "vol"
    heapsize = 5 * 1000 * 1000
    tentslab = TentSlab(mesh, method, heapsize)
    tentslab.SetMaxWavespeed(c)
    success = tentslab.PitchTents(dt, global_ct=global_ct)
    assert success, "Slab could not be pitched"
    maxslope = tentslab.MaxSlope()
    expected = 1.0 / c
    msg = "max slope {} exceeded {}".format(maxslope, expected)
    assert maxslope <= expected, msg


def test_1D_edge_causal():
    mesh = Get1DMesh()
    dt = 10
    c = 1
    global_ct = 0.999
    method = "edge"
    heapsize = 5 * 1000 * 1000
    tentslab = TentSlab(mesh, method, heapsize)
    tentslab.SetMaxWavespeed(c)
    success = tentslab.PitchTents(dt, global_ct=global_ct)
    assert success, "Slab could not be pitched"
    maxslope = tentslab.MaxSlope()
    expected = 1.0 / c
    msg = "max slope {} exceeded {}".format(maxslope, expected)
    assert maxslope <= expected, msg


def test_2D_vol_causal():
    mesh = Get2DMesh()
    dt = 10
    c = 1
    global_ct = 0.999
    method = "vol"
    heapsize = 5 * 1000 * 1000
    tentslab = TentSlab(mesh, method, heapsize)
    tentslab.SetMaxWavespeed(c)
    success = tentslab.PitchTents(dt, local_ct=True, global_ct=global_ct)
    assert success, "Slab could not be pitched"
    maxslope = tentslab.MaxSlope()
    expected = 1.0 / c
    msg = "max slope {} exceeded {}".format(maxslope, expected)
    assert maxslope <= expected, msg


def test_2D_edge_causal():
    mesh = Get2DMesh()
    dt = 10
    c = 1
    global_ct = 0.999
    method = "edge"
    heapsize = 5 * 1000 * 1000
    tentslab = TentSlab(mesh, method, heapsize)
    tentslab.SetMaxWavespeed(c)
    success = tentslab.PitchTents(dt, local_ct=True, global_ct=global_ct)
    assert success, "Slab could not be pitched"
    maxslope = tentslab.MaxSlope()
    expected = 1.0 / c
    msg = "max slope {} exceeded {}".format(maxslope, expected)
    assert maxslope <= expected, msg


def test_3D_vol_causal():
    mesh = Get2DMesh()
    dt = 10
    c = 1
    global_ct = 0.999
    method = "vol"
    heapsize = 5 * 1000 * 1000
    tentslab = TentSlab(mesh, method, heapsize)
    tentslab.SetMaxWavespeed(c)
    success = tentslab.PitchTents(dt, local_ct=True, global_ct=global_ct)
    assert success, "Slab could not be pitched"
    maxslope = tentslab.MaxSlope()
    expected = 1.0 / c
    msg = "max slope {} exceeded {}".format(maxslope, expected)
    assert maxslope <= expected, msg


def test_3D_edge_causal():
    mesh = Get2DMesh()
    dt = 10
    c = 1
    global_ct = 0.999
    method = "edge"
    heapsize = 5 * 1000 * 1000
    tentslab = TentSlab(mesh, method, heapsize)
    tentslab.SetMaxWavespeed(c)
    success = tentslab.PitchTents(dt, local_ct=True, global_ct=global_ct)
    assert success, "Slab could not be pitched"
    maxslope = tentslab.MaxSlope()
    expected = 1.0 / c
    msg = "max slope {} exceeded {}".format(maxslope, expected)
    assert maxslope <= expected, msg



"""
Select a 3D tents to visualize in pyplot
"""

from ngstents import TentSlab
from netgen.csg import unit_cube
import ngsolve as ng
import matplotlib.pyplot as plt

mesh = ng.Mesh(unit_cube.GenerateMesh(maxh=0.3))
print("creating tent pitched slab")
tps = TentSlab(mesh)
tps.SetMaxWavespeed(1.0)
tps.PitchTents(dt=0.5)
ntents = tps.GetNTents()
print("The slab has {} tents.".format(ntents))
resp = ""
plt.ion()
while True:
    resp = input("Enter a tent number in 0:{} or q to quit: ".format(ntents-1))
    plt.close()
    if resp.lower() == 'q':
        break
    try:
        n = int(resp)
    except ValueError:
        continue
    if 0 <= n < ntents:
        ax = tps.Draw3DTentPlt(n)
        plt.show()
    else:
        print("Invalid tent number")



from netgen.meshing import Mesh, FaceDescriptor, Element3D, Element2D, \
    MeshPoint, Pnt
import ngsolve as ng
import numpy as np


def DrawPitchedTents(self, uptolevel=None):  # self = TentSlab py object
    """
    Make a 2+1 dimensional mesh of spacetime tents, drawable using netgen.
    If uptolevel=L, then only tents of level L and lower are plotted,
    and if uptolevel is not given, all tents are included in the mesh.

    RETURNS:
      mesh3d: the spacetime mesh
      vertexlevels: tentlevels assigned to apex vertices (H1 function)
      tetlevels: tentlevels assigned to spacetime tetrahedra (DG function)
    """

    if self.mesh.dim != 2:
        raise NotImplementedError("Only supported for 2D spatial mesh")
    else:

        if self.mesh.ngmesh.GetNrIdentifications():
            raise NotImplementedError(
                "Don't know a good way to show periodic tents")

        # Make new a 3D mesh
        m = Mesh()
        fdbot = m.Add(FaceDescriptor(bc=1, domin=1, surfnr=1))
        fdwal = m.Add(FaceDescriptor(bc=2, domin=1, surfnr=2))
        fdtop = m.Add(FaceDescriptor(bc=3, domin=1, surfnr=3))
        bottomspatialmesh = self.mesh.ngmesh
        pmap = {}
        levelsnodal = {}

        for e in bottomspatialmesh.Elements2D():
            for v in e.vertices:
                if (v not in pmap):
                    pmap[v] = m.Add(bottomspatialmesh[v])
                    levelsnodal[v] = 0

        for e in bottomspatialmesh.Elements2D():
            m.Add(Element2D(fdbot, [pmap[v] for v in e.vertices]))

        # make:  tentsbylayer[layer] = list of tent nums in layer
        tentsbylayers = {i: [] for i in range(self.GetNLayers())}
        for i in range(self.GetNTents()):
            t = self.GetTent(i)
            tentsbylayers[t.level].append(i)

        bottomspatialelements = list(bottomspatialmesh.Elements2D())
        front = pmap
        levelstet = {}

        if uptolevel is None:
            uptolevel = self.GetNLayers()

        # Add surface elements and volume elements layer by layer
        for layer in range(uptolevel):
            for tn in tentsbylayers[layer]:
                t = self.GetTent(tn)
                pxy = self.mesh.vertices[t.vertex].point
                vtop = m.Add(MeshPoint(Pnt(pxy[0], pxy[1], t.ttop)))
                levelsnodal[vtop] = layer + 1
                vbot = front[t.vertex + 1]
                front[t.vertex + 1] = vtop

                # Make vertical (const in space) boundary elements
                for e in self.mesh.Elements(ng.BND):
                    if e.vertices[0].nr == t.vertex:
                        o = front[e.vertices[1].nr + 1]
                        m.Add(Element2D(fdwal, [vtop, vbot, o]))
                        # implicitly used orientation of e in above ordering
                    if e.vertices[1].nr == t.vertex:
                        o = front[e.vertices[0].nr + 1]
                        m.Add(Element2D(fdwal, [vtop, o, vbot]))

                # Make advancing front faces and new elements underneath
                for el in t.els:
                    topface = [
                        front[v] for v in bottomspatialelements[el].vertices
                    ]
                    m.Add(Element2D(fdtop, topface))
                    tet = m.Add(Element3D(1, topface + [vbot]))
                    levelstet[tet] = layer + 1

        m.Update()
        mesh3d = ng.Mesh(m)
        W = ng.L2(mesh3d, order=0)
        tetlevels = ng.GridFunction(W)
        tetlevels.vec.FV().NumPy()[:] = np.array(
            [levelstet[tet] for tet in levelstet])
        V = ng.H1(mesh3d, order=1)
        vertexlevels = ng.GridFunction(V)
        vertexlevels.vec.FV().NumPy()[:] = np.array(
            [levelsnodal[v] for v in levelsnodal])

        return mesh3d, vertexlevels, tetlevels



"""
Solve the Burgers equation in 1-D
(Use 1D Netgen Gui:  netgen burgers1d.py)
"""

from ngstents import TentSlab
from ngstents.conslaw import Burgers
from ngsolve.meshes import Make1DMesh
from ngsolve import x, exp, Draw, Redraw
from ngsolve import L2, GridFunction, TaskManager
from ngsolve.internal import viewoptions

step_sol = False  # Step through the solution

dt = 0.05  # tent slab height
c = 4.0  # maximal wave speed
p = 4  # spatial polynomial degree of approximation

mesh = Make1DMesh(100)

ts = TentSlab(mesh)
ts.SetMaxWavespeed(c)
ts.PitchTents(dt)
print("Max Slope: ", ts.MaxSlope())
#  ts.DrawPitchedTentsPlt()

V = L2(mesh, order=p)
u = GridFunction(V, "u")
burg = Burgers(u,
               ts,
               inflow=mesh.Boundaries("left"),
               outflow=mesh.Boundaries("right"))
burg.SetTentSolver(substeps=p * p)
burg.SetInitial(0.5 * exp(-100 * (x - 0.2) * (x - 0.2)))
Draw(u)

tend = 4
t = 0
cnt = 0

viewoptions.drawedges = 1
viewoptions.drawcolorbar = 0

input('start')
with TaskManager():
    while t < tend:
        burg.Propagate()
        t += dt
        cnt += 1
        if cnt % 5 == 0:
            print("{0:.5f}".format(t))
            Redraw()
            if step_sol:
                input('step')



from netgen.geom2d import unit_square
from ngsolve import Mesh, NodeId, FACET, VERTEX
from ngstents import TentSlab
import numpy as np


def test_tent_height():
    """
    Check that the edge slopes do not exceed 1/wavespeed.
    Passing this test does NOT imply the fulfillment of causality
    conditions.
    """

    mesh = Mesh(unit_square.GenerateMesh(maxh=.5))
    nref = 2
    for i in range(nref):
        mesh.Refine()
    dt = 0.05
    c = 16
    tol = 1e-12
    method = "vol"
    # Tent slab tests
    tentslab = TentSlab(mesh, method, 10**7)
    tentslab.SetMaxWavespeed(c)
    assert tentslab.PitchTents(dt), "Slab could not be pitched"
    for i in range(tentslab.GetNTents()):
        tent = tentslab.GetTent(i)
        tent_v = tent.vertex
        tent_pt = mesh[NodeId(VERTEX, tent_v)].point
        time_center = tent.ttop
        tent_facets = tent.internal_facets
        for edge_nr in tent_facets:
            edge = mesh[NodeId(FACET, edge_nr)]
            if edge.vertices[1].nr == tent_v:
                edge_v, other_v = edge.vertices[0].nr, edge.vertices[1].nr
            else:
                edge_v, other_v = edge.vertices[1].nr, edge.vertices[0].nr
            assert tent_v == other_v, "Error in tent structure"
            edge_pt = mesh[NodeId(VERTEX, edge_v)].point
            edge_pt_local_id = np.where(np.array(tent.nbv) == edge_v)[0][0]
            dist = np.linalg.norm(np.subtract(tent_pt, edge_pt))
            time_node = tent.nbtime[edge_pt_local_id]
            diff_t = time_center - time_node
            assert dist/c >= diff_t - tol, "Tent slope exceeds 1/c along edge!"



import matplotlib.pyplot as plt
import matplotlib.colors as mcolors


def DrawPitchedTentsPlt(self, showtentnums=False):
    """
    Draw 1D tents using Matplotlib
    """
    if self.mesh.dim != 1:
        raise NotImplementedError("Only supported for 1D spatial mesh")
    else:
        tents = self.TentData1D()
        pnts = []
        fig, ax = plt.subplots()

        colors = list(mcolors.BASE_COLORS)[:-1]  # remove white
        colors += list(mcolors.TABLEAU_COLORS)[:-1]  # omit tab:cyan close to c
        for pnt in self.mesh.ngmesh.Points():
            pnts.append(pnt.p[0])
        for i, tent in enumerate(tents):
            layer = tent[0][3] % len(colors)
            if len(tent) == 3:
                xvals = [pnts[tent[1][0]], pnts[tent[0][0]], pnts[tent[2][0]]]
                tvals = [tent[1][1], tent[0][1], tent[2][1]]
                xpos = pnts[tent[0][0]]
                tpos = 1 / 3 * (tent[1][1] + tent[0][1] + tent[2][1])
            else:
                xvals = [pnts[tent[1][0]], pnts[tent[0][0]]]
                tvals = [tent[1][1], tent[0][1]]
                xpos = 0.5 * (pnts[tent[1][0]] + pnts[tent[0][0]])
                tpos = 0.5 * (tent[1][1] + tent[0][1])
            ax.plot(xvals, tvals, color=colors[layer])
            if showtentnums:
                ax.text(xpos,
                        tpos,
                        str(i),
                        horizontalalignment='center',
                        verticalalignment='center')

        ax.set_ylabel('time ($t$)')
        ax.set_xlabel('space ($x$)')
        ax.margins(x=0, y=0)
        ax.axis('scaled')
        plt.show()



from ngsolve import Mesh, Draw, Redraw
from ngsolve import CoefficientFunction, exp, sqrt, sin, cos, x, y, z
from ngsolve import L2, GridFunction
from ngsolve import TaskManager
from ngsolve.internal import visoptions
from ngstents import TentSlab
from ngstents.conslaw import Advection
import time

def Make2DPeriodicMesh(xint, yint, maxh):
    # generate periodic mesh for xint x yint = [xmin, xmax] x [ymin, ymax]
    from netgen.geom2d import SplineGeometry
    periodic = SplineGeometry()
    pnts = [(xint[0],yint[0]),(xint[1],yint[0]),(xint[1],yint[1]),(xint[0],yint[1])]
    pnums = [periodic.AppendPoint(*p) for p in pnts]
    lbot = periodic.Append ( ["line", pnums[0], pnums[1]], bc="bottom")
    lright = periodic.Append ( ["line", pnums[1], pnums[2]], bc="right")
    periodic.Append ( ["line", pnums[0], pnums[3]], leftdomain=0, rightdomain=1, bc="left", copy=lright)
    periodic.Append ( ["line", pnums[3], pnums[2]], leftdomain=0, rightdomain=1, bc="top", copy=lbot)
    return periodic.GenerateMesh(maxh=maxh)

maxh = 0.1
mesh = Mesh(Make2DPeriodicMesh([0,1], [0,1], maxh))

# setting the problem
tend = 10
dt = 0.2
wavespeed = 2

# using causality constant
local_ctau = True
global_ctau = 1
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

order = 4
V = L2(mesh, order=order)
u = GridFunction(V)
cl = Advection(u, ts)
flux = (1,0.1)
cl.SetVectorField( CoefficientFunction(flux) )
cl.SetTentSolver("SAT",stages=order+1, substeps=2*order)

pos = (0.5,0.5)
u0 = CoefficientFunction( exp(-100* ((x-pos[0])*(x-pos[0])+(y-pos[1])*(y-pos[1])) ))
cl.SetInitial(u0)

Draw(u)
visoptions.autoscale = 0

t = 0
cnt = 0
redraw = 1

import time
input("press enter to start")
t1 = time.time()
with TaskManager():
    while t < tend-dt/2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt%redraw == 0:
            print("{:5f}".format(t))
            Redraw(True)
print("total time = ",time.time()-t1)



'''
Solve Burgers equation: d_t(u) - div f(u) = 0

f(u) = 1/2 * u^2

'''

from ngsolve import Mesh, Draw, Redraw
from ngsolve import CoefficientFunction, x, cos, sqrt, exp, IfPos, Id, InnerProduct
from ngsolve import L2, GridFunction, TaskManager, SetNumThreads
from ngsolve import specialcf as scf
from ngsolve.internal import visoptions, viewoptions
from ngstents import TentSlab
from ngsolve.meshes import Make1DMesh
from ngstents.conslaw import ConservationLaw
from math import pi

N = 20
mesh = Make1DMesh(N)

tend = 4
dt = 0.05
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(4)
ts.PitchTents(dt=dt)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

order = 4
V = L2(mesh, order=order)
gfu = GridFunction(V, name="u")

n = scf.normal(mesh.dim)
h = scf.mesh_size


def Flux(u):
    """
    Compute the flux for given TrialFunction u,
    a matrix-valued CoefficientFunction of size (h,w) where
    h represents the number of equations and
    w should correspond to the dimension of the mesh/space
    """
    return CoefficientFunction(1 / 2 * u**2)


def NumFlux(um, up):
    """
    Compute the numeric flux
    Inputs
    um: trace of u for current element on facet
    up: trace of u for other element on facet
    """
    # upwind flux
    return 0.5 * IfPos((um + up) * n, um**2, up**2) * n


def InverseMap(y):
    """
    solves "y = u - (f(u),gradphi)" for u
    """
    return 2 * y / (1 + sqrt(1 - 2 * ts.gradphi * y))


def Entropy(u):
    """
    Compute the entropy E for given TrialFunction u
    """
    return CoefficientFunction(u**2)


def EntropyFlux(u):
    """
    Compute the entropy flux F for given TrialFunction u
    """
    return CoefficientFunction(1 / 3 * u**3)


def NumEntropyFlux(um, up):
    """
    Compute the numerical flux of the entropy flux F for given
    um: trace of u for current element on facet
    up: trace of u for other element on facet
    """
    umean = 0.5 * (um + up)
    return 1 / 3 * IfPos(umean**2 * n, um**3, up**3) * n


def ViscosityCoefficient(u, res):
    """
    Compute the viscosity coefficient for given TrialFunction u
    and entropy residual res
    """
    nu_entr = (h / order)**2 * IfPos(res, res, -res) / Entropy(u)
    nu_max = h / order * IfPos(u, u, -u)
    return IfPos(nu_max - nu_entr, nu_entr, nu_max)


cl = ConservationLaw(gfu,
                     ts,
                     flux=Flux,
                     numflux=NumFlux,
                     inversemap=InverseMap,
                     entropy=Entropy,
                     entropyflux=EntropyFlux,
                     numentropyflux=NumEntropyFlux,
                     visccoeff=ViscosityCoefficient,
                     compile=True)
cl.SetTentSolver("SARK", stages=3, substeps=order * order)

# set inital data
cf0 = CoefficientFunction(0.5 * exp(-100 * (x - 0.2) * (x - 0.2)))
cl.SetInitial(cf0)

# set boundary conditions
cl.SetBoundaryCF(
    mesh.BoundaryCF({
        "left": NumFlux(cl.u_minus, cf0),
        "right": NumFlux(cl.u_minus, cl.u_minus)
    }))
cl.SetNumEntropyFlux(mesh.BoundaryCF({".*": EntropyFlux(cl.u_minus)}))

Draw(gfu)
viewoptions.drawedges = 1

redraw = 5
t = 0
cnt = 0
input("start")
import time

t1 = time.time()
with TaskManager():
    while t < tend - dt / 2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt % redraw == 0:
            print("{:5f}".format(t))
            Redraw(True)
print("total time = ", time.time() - t1)



from netgen.geom2d import SplineGeometry, unit_square
from ngsolve import Mesh, Draw, Redraw
from ngsolve import (CoefficientFunction, IfPos, sqrt, sin, cos, exp, log, x, y, z,
                     InnerProduct, Norm, OuterProduct, Id)
from ngsolve import L2, GridFunction, TaskManager, Timers, SetNumThreads, Integrate
from ngsolve import specialcf as scf
from ngsolve.internal import visoptions, viewoptions
from ngstents import TentSlab
from ngstents.conslaw import ConservationLaw
from ngstents.utils import Make1DMesh
from math import pi
import time

# define geometry
maxh = 0.05
mesh = Mesh(unit_square.GenerateMesh(maxh=maxh))

'''
solve euler equation:

d_t(g(u)) - div f(u) = 0

for u = [rho, m, E]^t
g(u) = u
f(u) = [[ m^t ],
        [ p*I + m^t*m/rho ],
        [ (E + p)*m^t/rho]]

p = 1/2*rho*T
T = 4/d (E/rho - 1/2 ||m||^2/rho^2)
degrees of freedom of the gas particle d = 5
'''

tend = 0.25
dt = 0.05

# using causality constant
local_ctau = True
global_ctau = 2/3
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(6)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

# vector-valued coefficient function
n = scf.normal(mesh.dim)
h = scf.mesh_size

def Flux(u):
    """
    Compute the flux f(u) for given TrialFunction u,
    a matrix-valued CoefficientFunction of size (h,w) where
    h represents the number of equations and
    w should correspond to the dimension of the mesh/space
    """
    m = u[1:mesh.dim+1]
    p = 2/5 * (u[mesh.dim+1] - 1/2 * InnerProduct(m,m)/u[0])
    return CoefficientFunction((m,
                                Id(mesh.dim)*p + OuterProduct(m,m)/u[0],
                                (u[mesh.dim+1] + p)/u[0] * m),
                               dims=(V.dim, mesh.dim))

def NumFlux(um, up):
    """
    Compute the numeric flux
    Inputs
    um: trace of u for current element on facet
    up: trace of u for other element on facet
    """
    # um.dims = mesh.dim+2
    # upwind flux
    flux = 0.5 * (Flux(um) + Flux(up))*n + (um - up)
    return flux

def InverseMap(y):
    """
    solves "y = u - (f(u),gradphi)" for u
    
    [derived in disseration "Mapped Tent Pitching Schemes
    for Hyperbolic Systems" by C. Wintersteiger (2020)]
    """
    y_rho = y[0]
    y_m = y[1:mesh.dim+1]
    y_E = y[mesh.dim+1]
    d = 5
    a1 = d/2 * ( y_rho - InnerProduct(y_m, ts.gradphi) )
    a2 = 2*y_E*y_rho - InnerProduct(y_m,y_m)
    normgrad = InnerProduct(ts.gradphi,ts.gradphi)
    p = a2 / (a1 + sqrt( a1**2 - (d+1)*normgrad*a2))
    rho = y_rho**2 / (y_rho - (InnerProduct(y_m, ts.gradphi) + p*normgrad))
    m = rho/y_rho * (y_m + p*ts.gradphi)
    E = 1/y_rho * (rho*y_E + p*InnerProduct(m,ts.gradphi))
    return CoefficientFunction((rho, m, E))

def BndNumFlux(um):
    """
    defines numerical flux on boundary elements using the normal vector n
    um: trace of u for current element on facet
    """
    return CoefficientFunction(( um[mesh.dim]*n, 0))

def Entropy(u):
    """
    Compute the entropy E for given TrialFunction u
    """
    d = 5
    rho = u[0]
    m = u[1:mesh.dim+1]
    T = 4/d * (u[mesh.dim+1]/rho - 1/2 * InnerProduct(m,m)/rho**2)
    T = IfPos(-T, T, 1e-10)
    return rho * ( log(rho) - d/2 * log(T) )

def EntropyFlux(u):
    """
    Compute the entropy flux F for given TrialFunction u
    """
    return u[1:mesh.dim+1]/u[0] * Entropy(u)

def NumEntropyFlux(um, up):
    """
    Compute the numerical flux of the entropy flux F for given
    um: trace of u for current element on facet
    up: trace of u for other element on facet
    """
    mn = InnerProduct(um[1:mesh.dim+1],n)
    return IfPos(mn, mn/um[0] * Entropy(um), InnerProduct(up[1:mesh.dim+1],n)/up[0] * Entropy(up))

def ViscosityCoefficient(u, res):
    """
    Compute the viscosity coefficient for given TrialFunction u
    and entropy residual res
    """
    nu_entr = (h/order)**2 * IfPos(res, res, -res)
    d = 5
    rho = u[0]
    ip_m = InnerProduct(u[1:mesh.dim+1],u[1:mesh.dim+1])
    T = 4/d * (u[mesh.dim+1]/rho - 1/2 * ip_m/rho**2)
    nu_max = 1/20 * h/order * ( sqrt(ip_m) + rho * sqrt((d+2)/d * T) )
    return IfPos( nu_max - nu_entr, nu_entr, nu_max)

# define conservation law
order = 2
V = L2(mesh, order=order, dim=mesh.dim+2)
gfu = GridFunction(V,name="u")
cl = ConservationLaw(gfu, ts,
                     flux=Flux, numflux=NumFlux, inversemap=InverseMap,
                     entropy = Entropy, entropyflux = EntropyFlux,
                     numentropyflux = NumEntropyFlux, visccoeff = ViscosityCoefficient,
                     compile = True)
cl.SetTentSolver("SARK", substeps=2*order)

# set inital data
d = 5
rho = CoefficientFunction(0.1+exp(-200*((x-0.5)*(x-0.5)+(y-0.5)*(y-0.5))))
m = CoefficientFunction((0,0))
p = CoefficientFunction(0.1+exp(-200*((x-0.5)*(x-0.5)+(y-0.5)*(y-0.5))))
T = 2*p/rho
E = d/4*T*rho + 1/(2*rho)*m*m

cf = CoefficientFunction((rho,m,E))
cl.SetInitial(CoefficientFunction(cf))

cl.SetBoundaryCF(mesh.BoundaryCF({"left|bottom|right|top" : NumFlux(cl.u_minus, cl.u_minus)}))

Draw(gfu)
visoptions.scalfunction = "u:1"

t = 0
cnt = 0
redraw = 1

input("press enter to start")
t1 = time.time()
with TaskManager():
    while t < tend-dt/2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt%redraw == 0:
            print("{:5f}".format(t))
            Redraw(True)
print("total time = ",time.time()-t1)



from ngstents.conslaw import Maxwell
from ngstents import TentSlab
from ngsolve.internal import visoptions, viewoptions
from netgen.csg import CSGeometry, OrthoBrick, Pnt
from ngsolve import (Mesh, Draw, Redraw, CoefficientFunction, sqrt, sin, cos,
                     exp, x, y, z, TaskManager, L2, GridFunction)

geom = CSGeometry()
brick = OrthoBrick(Pnt(0, 0, 0), Pnt(1, 1, 1)).bc("reflect")
geom.Add(brick)
mesh = geom.GenerateMesh(maxh=0.1)
mesh = Mesh(mesh)

# setting the problem
t_end = 0.5
dt = 0.1
wavespeed = 1

# using causality constant
local_ctau = True
global_ctau = 1/2
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

# setting the approximation space order
order = 2
V = L2(mesh, order=order, dim=2*mesh.dim)
u = GridFunction(V,"u")
cl = Maxwell(u, ts, reflect=mesh.Boundaries("reflect"))
cl.SetTentSolver("SAT",stages=order+1, substeps=2*order)

# initial data
Ez = exp(-40*((x-0.5)*(x-0.5)+(y-0.5)*(y-0.5)+(z-0.5)*(z-0.5)))
E = CoefficientFunction((0,0,Ez))
H = CoefficientFunction((0,0,0))

cf = CoefficientFunction((E,H))
cl.SetInitial(cf)

cfE = CoefficientFunction((u[0],u[1],u[2]))
cfH = CoefficientFunction((u[3],u[4],u[5]))

Draw(cfE, mesh, name="E")
Draw(cfH, mesh, name="H")

visoptions.scalfunction = "E:3"
viewoptions.clipping.enable = 1
visoptions.clipsolution = 'scal'
viewoptions.clipping.ny = 0
viewoptions.clipping.nz = -1

redraw = 1
t = 0
cnt = 0
import time
t1 = time.time()
with TaskManager():
    while t < t_end-dt/2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt%redraw == 0:
            print("{:5f}".format(t))
            Redraw(True)
print("total time = ",time.time()-t1)



from netgen.geom2d import SplineGeometry
from ngsolve import (Mesh, Draw, Redraw, CoefficientFunction, sqrt, x, y,
                     L2, GridFunction, TaskManager)
from ngsolve.internal import visoptions
from ngstents import TentSlab
from ngstents.conslaw import Euler

fine_mesh = False
geom = SplineGeometry()
if(fine_mesh):
    maxh = 0.04
    maxh_edge = 0.01
    pnts = [(0,0),(0.6,0,0.01),(0.6,0.2,0.003),(3,0.2),(3,1),(0,1)]
else:
    maxh = 0.1
    maxh_edge = maxh
    pnts = [(0,0),(0.6,0,0.03),(0.6,0.2,0.003),(3,0.2),(3,1),(0,1)]
bcs = ["inflow","reflect","reflect","reflect","outflow","reflect"]

pind = [geom.AddPoint(*pnt) for pnt in pnts]
for i in range(len(pind)):
    if(i==2):
        geom.Append(['line',pind[i-1],pind[i]],bc=bcs[i],maxh=maxh_edge)
    else:
        geom.Append(['line',pind[i-1],pind[i]],bc=bcs[i])

mesh = Mesh(geom.GenerateMesh(maxh=maxh,grading=0.15))

dt = 0.025
tend = 1

# using causality constant
local_ctau = True
global_ctau = 2/3
wavespeed = 6
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

order = 4
V = L2(mesh, order=order, dim=mesh.dim+2)
u = GridFunction(V,"u")
cl = Euler(u, ts,
           inflow=mesh.Boundaries("inflow"),
           outflow=mesh.Boundaries("outflow"),
           reflect=mesh.Boundaries("reflect"))
cl.SetTentSolver("SARK",substeps=order*order)

d = 5
rho0 = CoefficientFunction(1.4)
u0 = CoefficientFunction((3,0))
p0 = CoefficientFunction(1)
T0 = 2*p0/rho0
E0 = rho0*(d/4*T0 + 1/2*(u0*u0))

cf = CoefficientFunction((rho0,rho0*u0,E0))
cl.SetInitial(cf)

Draw(u)
visoptions.scalfunction = "u:1"

t = 0
cnt = 0
import time
t1 = time.time()
with TaskManager():
    while t < tend-dt/2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt%1 == 0:
            print("{:5f}".format(t))
            Redraw(True)
print(time.time()-t1)



from ngsolve import Mesh, Draw, Redraw
from ngsolve import CoefficientFunction, IfPos, exp, x, y, InnerProduct
from ngsolve import L2, GridFunction, TaskManager
from ngsolve import specialcf as scf
from ngsolve.internal import visoptions
from ngstents import TentSlab
from ngstents.conslaw import ConservationLaw
import time


def Make2DPeriodicMesh(xint, yint, maxh):
    # generate periodic mesh for xint x yint = [xmin, xmax] x [ymin, ymax]
    from netgen.geom2d import SplineGeometry
    periodic = SplineGeometry()
    pnts = [(xint[0], yint[0]), (xint[1], yint[0]),
            (xint[1], yint[1]), (xint[0], yint[1])]
    pnums = [periodic.AppendPoint(*p) for p in pnts]
    lbot = periodic.Append(["line", pnums[0], pnums[1]], bc="bottom")
    lright = periodic.Append(["line", pnums[1], pnums[2]], bc="right")
    periodic.Append(["line", pnums[0], pnums[3]], leftdomain=0,
                    rightdomain=1, bc="left", copy=lright)
    periodic.Append(["line", pnums[3], pnums[2]], leftdomain=0,
                    rightdomain=1, bc="top", copy=lbot)
    return periodic.GenerateMesh(maxh=maxh)


maxh = 0.1
mesh = Mesh(Make2DPeriodicMesh([0, 1], [0, 1], maxh))

# setting the problem
tend = 1
dt = 0.2
wavespeed = 2

# using causality constant
local_ctau = True
global_ctau = 1
ts = TentSlab(mesh, method="edge")
ts.SetMaxWavespeed(wavespeed)
ts.PitchTents(dt=dt, local_ct=local_ctau, global_ct=global_ctau)
print("max slope", ts.MaxSlope())
print("n tents", ts.GetNTents())

order = 4
V = L2(mesh, order=order)
u = GridFunction(V)

# vector-valued coefficient function
b = CoefficientFunction((1, 0.1))
n = scf.normal(mesh.dim)


def Flux(u):
    """
    Compute the flux for given TrialFunction u,
    a matrix-valued CoefficientFunction of size (h,w) where
    h represents the number of equations and
    w should correspond to the dimension of the mesh/space
    """
    return CoefficientFunction(b*u, dims=(V.dim, mesh.dim))


def NumFlux(um, up):
    """
    Compute the numeric flux
    Inputs
    um: trace of u for current element on facet
    up: trace of u for other element on facet
    """
    # upwind flux
    bn = b*n
    return IfPos(bn, bn*um, bn*up)


def InverseMap(y):
    """
    solves "y = u - (b*u,gradphi)" for u
    """
    return y/(1-InnerProduct(b, ts.gradphi))


cl = ConservationLaw(u, ts, flux=Flux, numflux=NumFlux, inversemap=InverseMap)

cl.SetTentSolver("SAT", stages=order+1, substeps=2*order)

pos = (0.5, 0.5)
u0 = CoefficientFunction(
    exp(-100 * ((x-pos[0])*(x-pos[0])+(y-pos[1])*(y-pos[1]))))
cl.SetInitial(u0)

Draw(u)
visoptions.autoscale = 0

t = 0
cnt = 0
redraw = 1

input("press enter to start")
t1 = time.time()
with TaskManager():
    while t < tend-dt/2:
        cl.Propagate()
        t += dt
        cnt += 1
        if cnt % redraw == 0:
            print("{:5f}".format(t))
            Redraw(True)
print("total time = ", time.time()-t1)
